@@ -1,9 +1,12 @@
 # expose jupyter_ai_magics ipython extension
+# DO NOT REMOVE.
 from jupyter_ai_magics import load_ipython_extension, unload_ipython_extension
 
-from ._version import __version__
+# expose jupyter_ai_magics providers
+# DO NOT REMOVE.
+from jupyter_ai_magics.providers import *
 
-# imports to expose entry points. DO NOT REMOVE.
+from ._version import __version__
 from .extension import AiExtension
 
 
