@@ -5,9 +5,10 @@ from jupyter_ai.models import HumanChatMessage
 from jupyter_ai_magics.providers import BaseProvider
 from langchain.chains import ConversationalRetrievalChain
 from langchain.memory import ConversationBufferWindowMemory
-from langchain.prompts import PromptTemplate
+from langchain_core.prompts import PromptTemplate
 
 from .base import BaseChatHandler, SlashCommandRoutingType
+from .learn import LearnChatHandler, Retriever
 
 PROMPT_TEMPLATE = """Given the following conversation and a follow up question, rephrase the follow up question to be a standalone question.
 
@@ -18,6 +19,16 @@ Standalone question:"""
 CONDENSE_PROMPT = PromptTemplate.from_template(PROMPT_TEMPLATE)
 
 
+class CustomLearnException(Exception):
+    """Exception raised when Jupyter AI's /ask command is used without the required /learn command."""
+
+    def __init__(self):
+        super().__init__(
+            "Jupyter AI's default /ask command requires the default /learn command. "
+            "If you are overriding /learn via the entry points API, be sure to also override or disable /ask."
+        )
+
+
 class AskChatHandler(BaseChatHandler):
     """Processes messages prefixed with /ask. This actor will
     send the message as input to a RetrieverQA chain, that
@@ -33,24 +44,41 @@ class AskChatHandler(BaseChatHandler):
 
     uses_llm = True
 
-    def __init__(self, retriever, *args, **kwargs):
+    def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-        self._retriever = retriever
         self.parser.prog = "/ask"
         self.parser.add_argument("query", nargs=argparse.REMAINDER)
+        # NOTE: devs should use self.retriever property to access the retriever
+        self._retriever = None  # Will be set when accessed
+
+    @property
+    def retriever(self):
+        """Lazy-loads and caches the retriever instance"""
+        if self._retriever is None:
+            learn_chat_handler = self.chat_handlers.get("/learn")
+            if not isinstance(learn_chat_handler, LearnChatHandler):
+                raise CustomLearnException()
+
+            self._retriever = Retriever(learn_chat_handler=learn_chat_handler)
+
+        return self._retriever  # Return cached instance
 
     def create_llm_chain(
         self, provider: Type[BaseProvider], provider_params: Dict[str, str]
     ):
-        model_parameters = self.get_model_parameters(provider, provider_params)
-        self.llm = provider(**provider_params, **model_parameters)
+        unified_parameters = {
+            **provider_params,
+            **(self.get_model_parameters(provider, provider_params)),
+        }
+        self.llm = provider(**unified_parameters)
         memory = ConversationBufferWindowMemory(
             memory_key="chat_history", return_messages=True, k=2
         )
+
         self.llm_chain = ConversationalRetrievalChain.from_llm(
             self.llm,
-            self._retriever,
+            self.retriever,
             memory=memory,
             condense_question_prompt=CONDENSE_PROMPT,
             verbose=False,
@@ -68,8 +96,14 @@ class AskChatHandler(BaseChatHandler):
         self.get_llm_chain()
 
         try:
-            result = await self.llm_chain.acall({"question": query})
-            response = result["answer"]
+            with self.pending("Searching learned documents", message):
+                assert self.llm_chain
+                # TODO: migrate this class to use a LCEL `Runnable` instead of
+                # `Chain`, then remove the below ignore comment.
+                result = await self.llm_chain.acall(  # type:ignore[attr-defined]
+                    {"question": query}
+                )
+                response = result["answer"]
             self.reply(response, message)
         except AssertionError as e:
             self.log.error(e)
