@@ -1,11 +1,11 @@
-from typing import List
-
-from jupyter_ai.models import ChatMessage, ClearMessage
+from jupyter_ai.models import ClearRequest
 
 from .base import BaseChatHandler, SlashCommandRoutingType
 
 
 class ClearChatHandler(BaseChatHandler):
+    """Clear the chat panel and show the help menu"""
+
     id = "clear"
     name = "Clear chat messages"
     help = "Clear the chat window"
@@ -17,10 +17,10 @@ class ClearChatHandler(BaseChatHandler):
         super().__init__(*args, **kwargs)
 
     async def process_message(self, _):
-        self._chat_history.clear()
+        # Clear chat by triggering `RootChatHandler.on_clear_request()`.
         for handler in self._root_chat_handlers.values():
             if not handler:
                 continue
 
-            handler.broadcast_message(ClearMessage())
+            handler.on_clear_request(ClearRequest())
             break
