@@ -1,3 +1,4 @@
+import ast
 import asyncio
 import os
 import time
@@ -12,9 +13,9 @@ from jupyter_ai_magics.providers import BaseProvider
 from langchain.chains import LLMChain
 from langchain.llms import BaseLLM
 from langchain.output_parsers import PydanticOutputParser
-from langchain.prompts import PromptTemplate
-from langchain.pydantic_v1 import BaseModel
 from langchain.schema.output_parser import BaseOutputParser
+from langchain_core.prompts import PromptTemplate
+from pydantic import BaseModel
 
 
 class OutlineSection(BaseModel):
@@ -54,7 +55,7 @@ async def generate_outline(description, llm=None, verbose=False):
     chain = NotebookOutlineChain.from_llm(llm=llm, parser=parser, verbose=verbose)
     outline = await chain.apredict(description=description)
     outline = parser.parse(outline)
-    return outline.dict()
+    return outline.model_dump()
 
 
 class CodeImproverChain(LLMChain):
@@ -198,6 +199,15 @@ async def afill_outline(outline, llm, verbose=False):
     await asyncio.gather(*all_coros)
 
 
+# Check if the content of the cell is python code or not
+def is_not_python_code(source: str) -> bool:
+    try:
+        ast.parse(source)
+    except:
+        return True
+    return False
+
+
 def create_notebook(outline):
     """Create an nbformat Notebook object for a notebook outline."""
     nbf = nbformat.v4
@@ -212,6 +222,26 @@ def create_notebook(outline):
         nb["cells"].append(nbf.new_markdown_cell("## " + section["title"]))
         for code_block in section["code"].split("\n\n"):
             nb["cells"].append(nbf.new_code_cell(code_block))
+
+    # Post process notebook for hanging code cells: merge hanging cell with the previous cell
+    merged_cells = []
+    for cell in nb["cells"]:
+        # Fix a hanging code cell
+        follows_code_cell = merged_cells and merged_cells[-1]["cell_type"] == "code"
+        is_incomplete = cell["cell_type"] == "code" and cell["source"].startswith(" ")
+        if follows_code_cell and is_incomplete:
+            merged_cells[-1]["source"] = (
+                merged_cells[-1]["source"] + "\n\n" + cell["source"]
+            )
+        else:
+            merged_cells.append(cell)
+
+    # Fix code cells that should be markdown
+    for cell in merged_cells:
+        if cell["cell_type"] == "code" and is_not_python_code(cell["source"]):
+            cell["cell_type"] = "markdown"
+
+    nb["cells"] = merged_cells
     return nb
 
 
@@ -223,15 +253,9 @@ class GenerateChatHandler(BaseChatHandler):
 
     uses_llm = True
 
-    def __init__(self, preferred_dir: str, log_dir: Optional[str], *args, **kwargs):
+    def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
-        self.log_dir = Path(log_dir) if log_dir else None
-        self.preferred_dir = (
-            os.path.abspath(os.path.expanduser(preferred_dir))
-            if preferred_dir != ""
-            else None
-        )
-        self.llm = None
+        self.llm: Optional[BaseProvider] = None
 
     def create_llm_chain(
         self, provider: Type[BaseProvider], provider_params: Dict[str, str]
@@ -253,6 +277,7 @@ class GenerateChatHandler(BaseChatHandler):
         # Save the user input prompt, the description property is now LLM generated.
         outline["prompt"] = prompt
 
+        assert self.llm
         if self.llm.allows_concurrency:
             # fill the outline concurrently
             await afill_outline(outline, llm=self.llm, verbose=True)
@@ -262,7 +287,7 @@ class GenerateChatHandler(BaseChatHandler):
 
         # create and write the notebook to disk
         notebook = create_notebook(outline)
-        final_path = os.path.join(self._output_dir, outline["title"] + ".ipynb")
+        final_path = os.path.join(self.output_dir, outline["title"] + ".ipynb")
         nbformat.write(notebook, final_path)
         return final_path
 
@@ -279,7 +304,7 @@ class GenerateChatHandler(BaseChatHandler):
 
     async def handle_exc(self, e: Exception, message: HumanChatMessage):
         timestamp = time.strftime("%Y-%m-%d-%H.%M.%S")
-        default_log_dir = Path(self._output_dir) / "jupyter-ai-logs"
+        default_log_dir = Path(self.output_dir) / "jupyter-ai-logs"
         log_dir = self.log_dir or default_log_dir
         log_dir.mkdir(parents=True, exist_ok=True)
         log_path = log_dir / f"generate-{timestamp}.log"
@@ -288,12 +313,3 @@ class GenerateChatHandler(BaseChatHandler):
 
         response = f"An error occurred while generating the notebook. The error details have been saved to `./{log_path}`.\n\nTry running `/generate` again, as some language models require multiple attempts before a notebook is generated."
         self.reply(response, message)
-
-    @property
-    def _output_dir(self):
-        # preferred dir is preferred, but if it is not specified,
-        # or if user removed it after startup, fallback to root.
-        if self.preferred_dir and os.path.exists(self.preferred_dir):
-            return self.preferred_dir
-        else:
-            return self.root_dir
