@@ -1,11 +1,11 @@
 import json
 import logging
 import os
-import shutil
 import time
-from typing import List, Optional, Union
+from copy import deepcopy
+from typing import List, Optional, Type, Union
 
-from deepmerge import always_merger as Merger
+from deepmerge import Merger, always_merger
 from jsonschema import Draft202012Validator as Validator
 from jupyter_ai.models import DescribeConfigResponse, GlobalConfig, UpdateConfigRequest
 from jupyter_ai_magics import JupyternautPersona, Persona
@@ -60,7 +60,7 @@ class BlockedModelError(Exception):
     pass
 
 
-def _validate_provider_authn(config: GlobalConfig, provider: AnyProvider):
+def _validate_provider_authn(config: GlobalConfig, provider: Type[AnyProvider]):
     # TODO: handle non-env auth strategies
     if not provider.auth_strategy or provider.auth_strategy.type != "env":
         return
@@ -97,16 +97,20 @@ class ConfigManager(Configurable):
         config=True,
     )
 
+    model_provider_id: Optional[str] = None
+    embeddings_provider_id: Optional[str] = None
+    completions_model_provider_id: Optional[str] = None
+
     def __init__(
         self,
         log: Logger,
         lm_providers: LmProvidersDict,
         em_providers: EmProvidersDict,
-        allowed_providers: Optional[List[str]],
-        blocked_providers: Optional[List[str]],
-        allowed_models: Optional[List[str]],
-        blocked_models: Optional[List[str]],
         defaults: dict,
+        allowed_providers: Optional[List[str]] = None,
+        blocked_providers: Optional[List[str]] = None,
+        allowed_models: Optional[List[str]] = None,
+        blocked_models: Optional[List[str]] = None,
         *args,
         **kwargs,
     ):
@@ -123,7 +127,13 @@ class ConfigManager(Configurable):
         self._allowed_models = allowed_models
         self._blocked_models = blocked_models
         self._defaults = defaults
-        """Provider defaults."""
+        """
+        Dictionary that maps config keys (e.g. `model_provider_id`, `fields`) to
+        user-specified overrides, set by traitlets configuration.
+
+        Values in this dictionary should never be mutated as they may refer to
+        entries in the global `self.settings` dictionary.
+        """
 
         self._last_read: Optional[int] = None
         """When the server last read the config file. If the file was not
@@ -139,11 +149,55 @@ class ConfigManager(Configurable):
         self._init_config()
 
     def _init_config_schema(self):
-        if not os.path.exists(self.schema_path):
-            os.makedirs(os.path.dirname(self.schema_path), exist_ok=True)
-            shutil.copy(OUR_SCHEMA_PATH, self.schema_path)
+        """
+        Initializes `config_schema.json` in the user's data dir whenever the
+        server extension starts. Users may add custom fields to their config
+        schema to insert new keys into the Jupyter AI config.
 
-    def _init_validator(self) -> Validator:
+        New in v2.31.1: Jupyter AI now merges the user's existing config schema
+        with Jupyter AI's config schema on init. This prevents validation errors
+        on missing keys when users upgrade Jupyter AI from an older version.
+
+        TODO v3: Remove the ability for users to provide a custom config schema.
+        This feature is entirely unused as far as I am aware, and we need to
+        simplify how Jupyter AI handles user configuration in v3 anyways.
+        """
+
+        # ensure the parent directory has been created
+        os.makedirs(os.path.dirname(self.schema_path), exist_ok=True)
+
+        # read existing_schema
+        if os.path.exists(self.schema_path):
+            with open(self.schema_path, encoding="utf-8") as f:
+                existing_schema = json.load(f)
+        else:
+            existing_schema = {}
+
+        # read default_schema
+        with open(OUR_SCHEMA_PATH, encoding="utf-8") as f:
+            default_schema = json.load(f)
+
+        # Create a custom `deepmerge.Merger` object to merge lists using the
+        # 'append_unique' strategy.
+        #
+        # This stops type union declarations like `["string", "null"]` from
+        # growing into `["string", "null", "string", "null"]` on restart.
+        # This fixes issue #1320.
+        merger = Merger(
+            [(list, ["append_unique"]), (dict, ["merge"]), (set, ["union"])],
+            ["override"],
+            ["override"],
+        )
+
+        # merge existing_schema into default_schema
+        # specifying existing_schema as the second argument ensures that
+        # existing_schema always overrides existing keys in default_schema, i.e.
+        # this call only adds new keys in default_schema.
+        schema = merger.merge(default_schema, existing_schema)
+        with open(self.schema_path, encoding="utf-8", mode="w") as f:
+            json.dump(schema, f, indent=self.indentation_depth)
+
+    def _init_validator(self) -> None:
         with open(OUR_SCHEMA_PATH, encoding="utf-8") as f:
             schema = json.loads(f.read())
             Validator.check_schema(schema)
@@ -151,7 +205,11 @@ class ConfigManager(Configurable):
 
     def _init_config(self):
         default_config = self._init_defaults()
-        if os.path.exists(self.config_path):
+        # if the config file exists, read from it and use our defaults to fill
+        # out any missing fields. otherwise, create a new config file from our
+        # defaults.
+        # the `st_size` check treats empty 0-byte config files as non-existent.
+        if os.path.exists(self.config_path) and os.stat(self.config_path).st_size != 0:
             self._process_existing_config(default_config)
         else:
             self._create_default_config(default_config)
@@ -159,46 +217,74 @@ class ConfigManager(Configurable):
     def _process_existing_config(self, default_config):
         with open(self.config_path, encoding="utf-8") as f:
             existing_config = json.loads(f.read())
-            merged_config = Merger.merge(
+            if "embeddings_fields" not in existing_config:
+                existing_config["embeddings_fields"] = {}
+            merged_config = always_merger.merge(
                 default_config,
                 {k: v for k, v in existing_config.items() if v is not None},
             )
             config = GlobalConfig(**merged_config)
-            validated_config = self._validate_lm_em_id(config)
+            validated_config = self._validate_model_ids(config)
 
             # re-write to the file to validate the config and apply any
             # updates to the config file immediately
             self._write_config(validated_config)
 
-    def _validate_lm_em_id(self, config):
-        lm_id = config.model_provider_id
-        em_id = config.embeddings_provider_id
+    def _validate_model_ids(self, config):
+        lm_provider_keys = ["model_provider_id", "completions_model_provider_id"]
+        em_provider_keys = ["embeddings_provider_id"]
+        clm_provider_keys = ["completions_model_provider_id"]
 
         # if the currently selected language or embedding model are
         # forbidden, set them to `None` and log a warning.
-        if lm_id is not None and not self._validate_model(lm_id, raise_exc=False):
-            self.log.warning(
-                f"Language model {lm_id} is forbidden by current allow/blocklists. Setting to None."
-            )
-            config.model_provider_id = None
-        if em_id is not None and not self._validate_model(em_id, raise_exc=False):
-            self.log.warning(
-                f"Embedding model {em_id} is forbidden by current allow/blocklists. Setting to None."
-            )
-            config.embeddings_provider_id = None
+        for lm_key in lm_provider_keys:
+            lm_id = getattr(config, lm_key)
+            if lm_id is not None and not self._validate_model(lm_id, raise_exc=False):
+                self.log.warning(
+                    f"Language model {lm_id} is forbidden by current allow/blocklists. Setting to None."
+                )
+                setattr(config, lm_key, None)
+        for em_key in em_provider_keys:
+            em_id = getattr(config, em_key)
+            if em_id is not None and not self._validate_model(em_id, raise_exc=False):
+                self.log.warning(
+                    f"Embedding model {em_id} is forbidden by current allow/blocklists. Setting to None."
+                )
+                setattr(config, em_key, None)
+        for clm_key in clm_provider_keys:
+            clm_id = getattr(config, clm_key)
+            if clm_id is not None and not self._validate_model(clm_id, raise_exc=False):
+                self.log.warning(
+                    f"Completion model {clm_id} is forbidden by current allow/blocklists. Setting to None."
+                )
+                setattr(config, clm_key, None)
 
         # if the currently selected language or embedding model ids are
         # not associated with models, set them to `None` and log a warning.
-        if lm_id is not None and not get_lm_provider(lm_id, self._lm_providers)[1]:
-            self.log.warning(
-                f"No language model is associated with '{lm_id}'. Setting to None."
-            )
-            config.model_provider_id = None
-        if em_id is not None and not get_em_provider(em_id, self._em_providers)[1]:
-            self.log.warning(
-                f"No embedding model is associated with '{em_id}'. Setting to None."
-            )
-            config.embeddings_provider_id = None
+        for lm_key in lm_provider_keys:
+            lm_id = getattr(config, lm_key)
+            if lm_id is not None and not get_lm_provider(lm_id, self._lm_providers)[1]:
+                self.log.warning(
+                    f"No language model is associated with '{lm_id}'. Setting to None."
+                )
+                setattr(config, lm_key, None)
+        for em_key in em_provider_keys:
+            em_id = getattr(config, em_key)
+            if em_id is not None and not get_em_provider(em_id, self._em_providers)[1]:
+                self.log.warning(
+                    f"No embedding model is associated with '{em_id}'. Setting to None."
+                )
+                setattr(config, em_key, None)
+        for clm_key in clm_provider_keys:
+            clm_id = getattr(config, clm_key)
+            if (
+                clm_id is not None
+                and not get_lm_provider(clm_id, self._lm_providers)[1]
+            ):
+                self.log.warning(
+                    f"No completion model is associated with '{clm_id}'. Setting to None."
+                )
+                setattr(config, clm_key, None)
 
         return config
 
@@ -206,19 +292,23 @@ class ConfigManager(Configurable):
         self._write_config(GlobalConfig(**default_config))
 
     def _init_defaults(self):
-        field_list = GlobalConfig.__fields__.keys()
-        properties = self.validator.schema.get("properties", {})
-        field_dict = {
-            field: properties.get(field).get("default") for field in field_list
+        config_keys = GlobalConfig.model_fields.keys()
+        schema_properties = self.validator.schema.get("properties", {})
+        default_config = {
+            field: schema_properties.get(field, {}).get("default")
+            for field in config_keys
         }
         if self._defaults is None:
-            return field_dict
+            return default_config
 
-        for field in field_list:
-            default_value = self._defaults.get(field)
+        for config_key in config_keys:
+            # we call `deepcopy()` here to avoid directly referring to the
+            # values in `self._defaults`, as they map to entries in the global
+            # `self.settings` dictionary and may be mutated otherwise.
+            default_value = deepcopy(self._defaults.get(config_key))
             if default_value is not None:
-                field_dict[field] = default_value
-        return field_dict
+                default_config[config_key] = default_value
+        return default_config
 
     def _read_config(self) -> GlobalConfig:
         """Returns the user's current configuration as a GlobalConfig object.
@@ -232,6 +322,8 @@ class ConfigManager(Configurable):
         with open(self.config_path, encoding="utf-8") as f:
             self._last_read = time.time_ns()
             raw_config = json.loads(f.read())
+            if "embeddings_fields" not in raw_config:
+                raw_config["embeddings_fields"] = {}
             config = GlobalConfig(**raw_config)
             self._validate_config(config)
             return config
@@ -241,7 +333,7 @@ class ConfigManager(Configurable):
         read and before every write to the config file. Guarantees that the
         config file conforms to the JSON Schema, and that the language and
         embedding models have authn credentials if specified."""
-        self.validator.validate(config.dict())
+        self.validator.validate(config.model_dump())
 
         # validate language model config
         if config.model_provider_id:
@@ -261,6 +353,36 @@ class ConfigManager(Configurable):
             # verify model is authenticated
             _validate_provider_authn(config, lm_provider)
 
+            # verify fields exist for this model if needed
+            if lm_provider.fields and config.model_provider_id not in config.fields:
+                config.fields[config.model_provider_id] = {}
+
+        # validate completions model config
+        if config.completions_model_provider_id:
+            _, completions_provider = get_lm_provider(
+                config.completions_model_provider_id, self._lm_providers
+            )
+
+            # verify model is declared by some provider
+            if not completions_provider:
+                raise ValueError(
+                    f"No language model is associated with '{config.completions_model_provider_id}'."
+                )
+
+            # verify model is not blocked
+            self._validate_model(config.completions_model_provider_id)
+
+            # verify model is authenticated
+            _validate_provider_authn(config, completions_provider)
+
+            # verify completions fields exist for this model if needed
+            if (
+                completions_provider.fields
+                and config.completions_model_provider_id
+                not in config.completions_fields
+            ):
+                config.completions_fields[config.completions_model_provider_id] = {}
+
         # validate embedding model config
         if config.embeddings_provider_id:
             _, em_provider = get_em_provider(
@@ -279,6 +401,13 @@ class ConfigManager(Configurable):
             # verify model is authenticated
             _validate_provider_authn(config, em_provider)
 
+            # verify embedding fields exist for this model if needed
+            if (
+                em_provider.fields
+                and config.embeddings_provider_id not in config.embeddings_fields
+            ):
+                config.embeddings_fields[config.embeddings_provider_id] = {}
+
     def _validate_model(self, model_id: str, raise_exc=True):
         """
         Validates a model against the set of allow/blocklists specified by the
@@ -303,10 +432,13 @@ class ConfigManager(Configurable):
                     "Model provider included in the provider blocklist."
                 )
 
-            if self._allowed_models and model_id not in self._allowed_models:
+            if (
+                self._allowed_models is not None
+                and model_id not in self._allowed_models
+            ):
                 raise BlockedModelError("Model not included in the model allowlist.")
 
-            if self._blocked_models and model_id in self._blocked_models:
+            if self._blocked_models is not None and model_id in self._blocked_models:
                 raise BlockedModelError("Model included in the model blocklist.")
         except BlockedModelError as e:
             if raise_exc:
@@ -321,28 +453,31 @@ class ConfigManager(Configurable):
         complete `GlobalConfig` object, and should not be called publicly."""
         # remove any empty field dictionaries
         new_config.fields = {k: v for k, v in new_config.fields.items() if v}
+        new_config.completions_fields = {
+            k: v for k, v in new_config.completions_fields.items() if v
+        }
+        new_config.embeddings_fields = {
+            k: v for k, v in new_config.embeddings_fields.items() if v
+        }
 
         self._validate_config(new_config)
         with open(self.config_path, "w") as f:
-            json.dump(new_config.dict(), f, indent=self.indentation_depth)
+            json.dump(new_config.model_dump(), f, indent=self.indentation_depth)
 
     def delete_api_key(self, key_name: str):
-        config_dict = self._read_config().dict()
-        lm_provider = self.lm_provider
-        em_provider = self.em_provider
+        config_dict = self._read_config().model_dump()
         required_keys = []
-        if (
-            lm_provider
-            and lm_provider.auth_strategy
-            and lm_provider.auth_strategy.type == "env"
-        ):
-            required_keys.append(lm_provider.auth_strategy.name)
-        if (
-            em_provider
-            and em_provider.auth_strategy
-            and em_provider.auth_strategy.type == "env"
-        ):
-            required_keys.append(self.em_provider.auth_strategy.name)
+        for provider in [
+            self.lm_provider,
+            self.em_provider,
+            self.completions_lm_provider,
+        ]:
+            if (
+                provider
+                and provider.auth_strategy
+                and provider.auth_strategy.type == "env"
+            ):
+                required_keys.append(provider.auth_strategy.name)
 
         if key_name in required_keys:
             raise KeyInUseError(
@@ -352,7 +487,7 @@ class ConfigManager(Configurable):
         config_dict["api_keys"].pop(key_name, None)
         self._write_config(GlobalConfig(**config_dict))
 
-    def update_config(self, config_update: UpdateConfigRequest):
+    def update_config(self, config_update: UpdateConfigRequest):  # type:ignore
         last_write = os.stat(self.config_path).st_mtime_ns
         if config_update.last_read and config_update.last_read < last_write:
             raise WriteConflictError(
@@ -364,15 +499,15 @@ class ConfigManager(Configurable):
                 if not api_key_value:
                     raise KeyEmptyError("API key value cannot be empty.")
 
-        config_dict = self._read_config().dict()
-        Merger.merge(config_dict, config_update.dict(exclude_unset=True))
+        config_dict = self._read_config().model_dump()
+        always_merger.merge(config_dict, config_update.model_dump(exclude_unset=True))
         self._write_config(GlobalConfig(**config_dict))
 
     # this cannot be a property, as the parent Configurable already defines the
     # self.config attr.
     def get_config(self):
         config = self._read_config()
-        config_dict = config.dict(exclude_unset=True)
+        config_dict = config.model_dump(exclude_unset=True)
         api_key_names = list(config_dict.pop("api_keys").keys())
         return DescribeConfigResponse(
             **config_dict, api_keys=api_key_names, last_read=self._last_read
@@ -390,67 +525,84 @@ class ConfigManager(Configurable):
 
     @property
     def lm_provider(self):
-        config = self._read_config()
-        lm_gid = config.model_provider_id
-        if lm_gid is None:
-            return None
-
-        _, Provider = get_lm_provider(config.model_provider_id, self._lm_providers)
-        return Provider
+        return self._get_provider("model_provider_id", self._lm_providers)
 
     @property
     def em_provider(self):
+        return self._get_provider("embeddings_provider_id", self._em_providers)
+
+    @property
+    def completions_lm_provider(self):
+        return self._get_provider("completions_model_provider_id", self._lm_providers)
+
+    def _get_provider(self, key, listing):
         config = self._read_config()
-        em_gid = config.embeddings_provider_id
-        if em_gid is None:
+        gid = getattr(config, key)
+        if gid is None:
             return None
 
-        _, Provider = get_em_provider(em_gid, self._em_providers)
+        _, Provider = get_lm_provider(gid, listing)
         return Provider
 
     @property
     def lm_provider_params(self):
-        # get generic fields
-        config = self._read_config()
-        lm_gid = config.model_provider_id
-        if not lm_gid:
-            return None
-
-        lm_lid = lm_gid.split(":", 1)[1]
-        fields = config.fields.get(lm_gid, {})
-
-        # get authn fields
-        _, Provider = get_lm_provider(lm_gid, self._lm_providers)
-        authn_fields = {}
-        if Provider.auth_strategy and Provider.auth_strategy.type == "env":
-            key_name = Provider.auth_strategy.name
-            authn_fields[key_name.lower()] = config.api_keys[key_name]
-
-        return {
-            "model_id": lm_lid,
-            **fields,
-            **authn_fields,
-        }
+        return self._provider_params("model_provider_id", self._lm_providers)
 
     @property
     def em_provider_params(self):
-        # get generic fields
-        config = self._read_config()
-        em_gid = config.embeddings_provider_id
-        if not em_gid:
-            return None
+        return self._provider_params("embeddings_provider_id", self._em_providers)
 
-        em_lid = em_gid.split(":", 1)[1]
+    @property
+    def completions_lm_provider_params(self):
+        return self._provider_params(
+            "completions_model_provider_id", self._lm_providers, completions=True
+        )
+
+    def _provider_params(self, key, listing, completions: bool = False):
+        # read config
+        config = self._read_config()
+
+        # get model ID (without provider ID component) from model universal ID
+        # (with provider component).
+        model_uid = getattr(config, key)
+        if not model_uid:
+            return None
+        model_id = model_uid.split(":", 1)[1]
+
+        # get config fields (e.g. base API URL, etc.)
+        if completions:
+            fields = config.completions_fields.get(model_uid, {})
+        elif key == "embeddings_provider_id":
+            fields = config.embeddings_fields.get(model_uid, {})
+        else:
+            fields = config.fields.get(model_uid, {})
+
+        # exclude empty fields
+        # TODO: modify the config manager to never save empty fields in the
+        # first place.
+        fields = {
+            k: None if isinstance(v, str) and not len(v) else v
+            for k, v in fields.items()
+        }
 
         # get authn fields
-        _, Provider = get_em_provider(em_gid, self._em_providers)
+        _, Provider = (
+            get_em_provider(model_uid, listing)
+            if key == "embeddings_provider_id"
+            else get_lm_provider(model_uid, listing)
+        )
         authn_fields = {}
         if Provider.auth_strategy and Provider.auth_strategy.type == "env":
+            keyword_param = (
+                Provider.auth_strategy.keyword_param
+                or Provider.auth_strategy.name.lower()
+            )
             key_name = Provider.auth_strategy.name
-            authn_fields[key_name.lower()] = config.api_keys[key_name]
+            authn_fields[keyword_param] = config.api_keys[key_name]
 
         return {
-            "model_id": em_lid,
+            "model_id": model_id,
+            **fields,
             **authn_fields,
         }
 
