@@ -1,13 +1,20 @@
 import { useState, useEffect, useMemo, useCallback } from 'react';
 import { AiService } from '../../handler';
+import { getProviderId, getModelLocalId } from '../../utils';
 
-type ServerInfoProperties = {
-  config: AiService.DescribeConfigResponse;
-  lmProviders: AiService.ListProvidersResponse;
-  emProviders: AiService.ListProvidersResponse;
+type ProvidersInfo = {
   lmProvider: AiService.ListProvidersEntry | null;
   emProvider: AiService.ListProvidersEntry | null;
   lmLocalId: string;
+  emLocalId: string;
+};
+
+type ServerInfoProperties = {
+  lmProviders: AiService.ListProvidersResponse;
+  emProviders: AiService.ListProvidersResponse;
+  config: AiService.DescribeConfigResponse;
+  chat: ProvidersInfo;
+  completions: Omit<ProvidersInfo, 'emProvider' | 'emLocalId'>;
 };
 
 type ServerInfoMethods = {
@@ -63,14 +70,28 @@ export function useServerInfo(): ServerInfo {
         lmGid === null ? null : getProvider(lmGid, lmProviders);
       const emProvider =
         emGid === null ? null : getProvider(emGid, emProviders);
-      const lmLocalId = lmGid === null ? '' : getLocalId(lmGid);
+      const lmLocalId = (lmGid && getModelLocalId(lmGid)) ?? '';
+      const emLocalId = (emGid && getModelLocalId(emGid)) ?? '';
+
+      const clmGid = config.completions_model_provider_id;
+      const clmProvider =
+        clmGid === null ? null : getProvider(clmGid, lmProviders);
+      const clmLocalId = (clmGid && getModelLocalId(clmGid)) ?? '';
+
       setServerInfoProps({
         config,
         lmProviders,
         emProviders,
-        lmProvider,
-        emProvider,
-        lmLocalId
+        chat: {
+          lmProvider,
+          emProvider,
+          lmLocalId,
+          emLocalId
+        },
+        completions: {
+          lmProvider: clmProvider,
+          lmLocalId: clmLocalId
+        }
       });
 
       setState(ServerInfoState.Ready);
@@ -135,25 +156,3 @@ function getProvider(
   const provider = providers.providers.find(p => p.id === providerId);
   return provider ?? null;
 }
-
-function getProviderId(gid: string) {
-  if (!gid) {
-    return null;
-  }
-
-  const components = gid.split(':');
-  if (components.length < 2) {
-    return null;
-  }
-
-  return components[0];
-}
-
-function getLocalId(gid: string) {
-  const components = gid.split(':');
-  if (components.length < 2) {
-    return '';
-  }
-
-  return components[1];
-}
